//
// Darkulator - BSP file brightness utility by botman@planethalflife.com
//
// Darkulator.cpp
//

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "bspfile.h"

bool isWindowsApp = TRUE;  // for ERROR reporting routine

char szFilter[] = {"BSP Files (*.bsp)|*.bsp|"};
char szLoadTitle[] = {"Darkulator - Select BSP file to load"};
char szSaveTitle[] = {"Darkulator - Save BSP file"};
char szFile[MAX_PATH], szFileTitle[MAX_PATH];

char szExplorerPath[MAX_PATH];
char szBSPfilename[MAX_PATH];
char szBSPpathname[MAX_PATH];

char skyname[64];
float brightness;

char	save_dentdata[MAX_MAP_ENTSTRING];
byte *save_dlightdata = NULL;


UINT GetPathFromFullPathName( LPCTSTR lpFullPathName, LPTSTR lpPathBuffer, int nPathBufferLength );
UINT CALLBACK CenterOpenDlgBox(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK DialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);


int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow)
{
   char szExePathName[MAX_PATH];  // full pathname of the application
   int index;
   HWND hWnd;
   MSG msg;

   InitCommonControls();  // needed for slider control

   GetModuleFileName( hInstance, szExePathName, sizeof(szExePathName) );
   GetPathFromFullPathName( szExePathName, szExplorerPath, sizeof(szExplorerPath) );

   for (index = 0; index < MAX_MAP_ENTITIES; index++)
      entities[index].epairs = NULL;

   skyname[0] = 0;
   szBSPfilename[0] = 0;

   for (index = 0; szFilter[index] != '\0'; index++)
      if (szFilter[index] == '|')
         szFilter[index] = '\0';

   hWnd = CreateDialog(hInstance, MAKEINTRESOURCE( IDD_DIALOG1 ), 0, DialogProc);

   if (!hWnd)
   {
      MessageBox(NULL, "Can't create Dialog", NULL, MB_OK);
      return 1;
   }

   while (GetMessage(&msg, NULL, 0, 0)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
   }

   return msg.wParam;
}


UINT GetPathFromFullPathName( LPCTSTR lpFullPathName, LPTSTR lpPathBuffer,
                              int nPathBufferLength )
{
   int nLength;
   int i;

   if ((nLength = lstrlen( lpFullPathName )) > nPathBufferLength)
      return( nLength );

   lstrcpy( lpPathBuffer, lpFullPathName );

   for( i = lstrlen( lpPathBuffer );
        (lpPathBuffer[i] != '\\') && (lpPathBuffer[i] != ':'); i-- )
      ;
   if (':' == lpPathBuffer[i])
      lpPathBuffer[i+1] = '\0';
   else
      lpPathBuffer[i] = '\0';

   nLength = i;

   for (i = 0; i < nLength; i++)
      lpPathBuffer[i] = tolower(lpPathBuffer[i]);

   return( (UINT) nLength );
}


BOOL CenterWindow (HWND hWnd)
{
    RECT    rRect, rParentRect;
    HWND    hParentWnd;
    int     wParent, hParent, xNew, yNew;
    int     w, h;

    GetWindowRect(hWnd, &rRect);
    w = rRect.right - rRect.left;
    h = rRect.bottom - rRect.top;

    if (NULL == (hParentWnd = GetParent( hWnd )))
       hParentWnd = GetDesktopWindow();

    GetWindowRect( hParentWnd, &rParentRect );

    wParent = rParentRect.right - rParentRect.left;
    hParent = rParentRect.bottom - rParentRect.top;

    xNew = wParent/2 - w/2 + rParentRect.left;
    yNew = hParent/2 - h/2 + rParentRect.top;

    return SetWindowPos(hWnd, NULL, xNew, yNew, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}


UINT CALLBACK CenterOpenDlgBox(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
   LPOFNOTIFY pNotify;

   switch (uMsg)
   {
      case WM_NOTIFY:
         pNotify = (LPOFNOTIFY)lParam;

         if (pNotify->hdr.code == CDN_INITDONE)
         {
            CenterWindow( GetParent(hWnd) );
            return( TRUE );
         }
   }

   return( FALSE );
}


void FreeGlobals(void)
{
   FreeEntities();

   if (dmodels)
   {
      free(dmodels);
      dmodels = NULL;
   }

   if (dvisdata)
   {
      free(dvisdata);
      dvisdata = NULL;
   }

   if (dlightdata)
   {
      free(dlightdata);
      dlightdata = NULL;
   }

   if (dtexdata)
   {
      free(dtexdata);
      dtexdata = NULL;
   }

   if (dleafs)
   {
      free(dleafs);
      dleafs = NULL;
   }

   if (dplanes)
   {
      free(dplanes);
      dplanes = NULL;
   }

   if (dvertexes)
   {
      free(dvertexes);
      dvertexes = NULL;
   }

   if (dnodes)
   {
      free(dnodes);
      dnodes = NULL;
   }

   if (texinfo)
   {
      free(texinfo);
      texinfo = NULL;
   }

   if (dfaces)
   {
      free(dfaces);
      dfaces = NULL;
   }

   if (dclipnodes)
   {
      free(dclipnodes);
      dclipnodes = NULL;
   }

   if (dedges)
   {
      free(dedges);
      dedges = NULL;
   }

   if (dmarksurfaces)
   {
      free(dmarksurfaces);
      dmarksurfaces = NULL;
   }

   if (dsurfedges)
   {
      free(dsurfedges);
      dsurfedges = NULL;
   }
}


void ReplaceSkyname(void)
{
   int ent_index = -1;  // negative one initializes search

   if ((ent_index = FindEntityByClassname(ent_index, "worldspawn")) != -1)
   {
      if (skyname[0] == 0)  // no skyname defined?
      {
         // remove "skyname" key if it exists...
         RemoveKey(&entities[ent_index], "skyname");
      }
      else
      {
         // replace (or add) skyname key/value...
         SetKeyValue(&entities[ent_index], "skyname", skyname);
      }
   }
}


void ScaleLightEntity(int ent_index)
{
   char value[MAX_VALUE];
   int r, g, b, v;

   // look for a "_light" key...
   strcpy(value, ValueForKey(&entities[ent_index], "_light"));

   if (value[0])
   {
      if (sscanf(value, "%d %d %d %d", &r, &g, &b, &v) == 1)
      {
         g = b = r;
         v = -1;
      }

      if (v == -1)  // was only RED given in key/value pair?
      {
         // scale just the red color and write the value...
         r = (int)(r * brightness);
         if (r < 0)
            r = 0;
         if (r > 255)
            r = 255;

         wsprintf(value, "%d", r);
      }
      else
      {
         // scale the intensity and write the value...
         v = (int)(v * brightness);
         if (v < 0)
            v = 0;
         if (v > 255)
            v = 255;

         wsprintf(value, "%d %d %d %d", r, g, b, v);
      }

      // ent_index is always one ahead of the current entity...
      SetKeyValue(&entities[ent_index], "_light", value);
   }
}


void UpdateLightEntities(void)
{
   int ent_index;

   ent_index = -1;  // negative one initializes search
   while ((ent_index = FindEntityByClassname(ent_index, "light")) != -1)
      ScaleLightEntity(ent_index);

   ent_index = -1;  // negative one initializes search
   while ((ent_index = FindEntityByClassname(ent_index, "light_spot")) != -1)
      ScaleLightEntity(ent_index);

   ent_index = -1;  // negative one initializes search
   while ((ent_index = FindEntityByClassname(ent_index, "light_environment")) != -1)
      ScaleLightEntity(ent_index);
}


void UpdateLightmaps()
{
   int index;
   int value;

   if (dlightdata == NULL)  // no lightmaps defined?
      return;

   for (index = 0; index < lightdatasize; index++)
   {
      value = dlightdata[index];
      value = (int)(value * brightness);
      if (value < 0)
         value = 0;
      if (value > 255)
         value = 255;
      dlightdata[index] = value;
   }
}


void SaveBSPFile(HWND hWnd)
{
   OPENFILENAME ofn;

   memset( &ofn, 0, sizeof( ofn ) );

   szFile[0] = '\0';

   ofn.lStructSize = sizeof( ofn );
   ofn.hwndOwner = hWnd;
   ofn.hInstance = NULL;
   ofn.lpstrTitle = szSaveTitle;
   ofn.lpstrFilter = szFilter;
   ofn.nFilterIndex = 1;
   ofn.lpstrFile = szFile;
   ofn.nMaxFile = sizeof(szFile);
   ofn.lpstrFileTitle = szFileTitle;
   ofn.nMaxFileTitle = sizeof(szFileTitle);
   ofn.lpstrInitialDir = szExplorerPath;
   ofn.Flags = OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY | OFN_EXPLORER |
               OFN_ENABLESIZING | OFN_ENABLEHOOK;
   ofn.lpfnHook = CenterOpenDlgBox;

   if (!GetSaveFileName( &ofn ))
      return;

   GetPathFromFullPathName( szFile, szExplorerPath, sizeof(szExplorerPath) );

   if (strstr(szFileTitle, ".") == NULL)
      strcat(szFile, ".bsp");

   WriteBSPFile(szFile);


   MessageBox(hWnd, "BSP file was saved.", "Wooo-Hooo!", MB_OK);
}

BOOL CALLBACK DialogProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   char szDialogTitle[MAX_PATH];
   int pos;
   int ent_index;
   char szPos[4];
   OPENFILENAME ofn;

   switch (message)
   {
      case WM_INITDIALOG:

         if (!CenterWindow( hWnd ))
            return( FALSE );

         SendDlgItemMessage(hWnd, IDC_SLIDER1, TBM_SETRANGE, (WPARAM)TRUE, (LPARAM)MAKELONG(0,20));
         SendDlgItemMessage(hWnd, IDC_SLIDER1, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)10);
         SendDlgItemMessage(hWnd, IDC_EDIT2, WM_SETTEXT, 0, (LPARAM)((LPCSTR)"1.0"));

         break;

      case WM_COMMAND:

         switch (LOWORD(wParam))
         {
            case IDCLOSE:

               FreeGlobals();

               PostQuitMessage(0);

               return TRUE;

            case IDC_BUTTON1:

               memset( &ofn, 0, sizeof( ofn ) );

               szFile[0] = '\0';

               ofn.lStructSize = sizeof( ofn );
               ofn.hwndOwner = hWnd;
               ofn.hInstance = NULL;
               ofn.lpstrTitle = szLoadTitle;
               ofn.lpstrFilter = szFilter;
               ofn.nFilterIndex = 1;
               ofn.lpstrFile = szFile;
               ofn.nMaxFile = sizeof(szFile);
               ofn.lpstrFileTitle = szFileTitle;
               ofn.nMaxFileTitle = sizeof(szFileTitle);
               ofn.lpstrInitialDir = szExplorerPath;
               ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_EXPLORER |
                           OFN_ENABLESIZING | OFN_ENABLEHOOK;
               ofn.lpfnHook = CenterOpenDlgBox;

               if (GetOpenFileName( &ofn ))
               {
                  FreeGlobals();

                  strcpy(szBSPfilename, szFileTitle);
                  strcpy(szBSPpathname, szFile);

                  GetPathFromFullPathName( szBSPpathname, szExplorerPath, sizeof(szExplorerPath) );

                  LoadBSPFile(szBSPpathname);

                  ParseEntities();

                  skyname[0] = 0;
                  ent_index = -1;  // negative one initializes search

                  if ((ent_index = FindEntityByClassname(ent_index, "worldspawn")) != -1)
                  {
                     // look for a worldspawn "skyname" key...
                     strcpy(skyname, ValueForKey(&entities[ent_index], "skyname"));

                     if (skyname[0] == 0)  // not found, use the "message" key if it exists...
                     {
                        strcpy(skyname, ValueForKey(&entities[ent_index], "message"));
                        if (strstr(skyname, " "))  // blanks aren't valid...
                           skyname[0] = 0;
                     }
                  }

                  strcpy(szDialogTitle, "Darkulator - ");
                  strcat(szDialogTitle, szBSPfilename);

                  SetWindowText(hWnd, szDialogTitle);

                  SendDlgItemMessage(hWnd, IDC_EDIT1, WM_SETTEXT, 0, (LPARAM)skyname);

                  SendDlgItemMessage(hWnd, IDC_SLIDER1, TBM_SETRANGE, (WPARAM)TRUE, (LPARAM)MAKELONG(0,20));
                  SendDlgItemMessage(hWnd, IDC_SLIDER1, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)10);
                  SendDlgItemMessage(hWnd, IDC_EDIT2, WM_SETTEXT, 0, (LPARAM)((LPCSTR)"1.0"));
               }

               return TRUE;

            case IDC_BUTTON2:

               if (szBSPfilename[0] == 0)
               {
                  MessageBox(hWnd, "You have to Load a BSP file first!", "Error", MB_OK);
                  return TRUE;
               }

               pos = SendDlgItemMessage(hWnd, IDC_SLIDER1, TBM_GETPOS, 0, (LPARAM)0);
               brightness = (float)pos / 10.0f;

               SendDlgItemMessage(hWnd, IDC_EDIT1, WM_GETTEXT, sizeof(skyname), (LPARAM)skyname);

               ReplaceSkyname();

               // save a copy of the original light values...
               memcpy(save_dentdata, dentdata, entdatasize);
               if (dlightdata != NULL)
               {
                  save_dlightdata = (byte *)malloc(lightdatasize);
                  memcpy(save_dlightdata, dlightdata, lightdatasize);
               }

               UpdateLightEntities();
               UpdateLightmaps();

               UnparseEntities();

               SaveBSPFile(hWnd);

               // restore the original light values...
               memcpy(dentdata, save_dentdata, entdatasize);
               if (dlightdata != NULL)
               {
                  memcpy(dlightdata, save_dlightdata, lightdatasize);
                  free(save_dlightdata);
               }

               return TRUE;
         }

         break;

      case WM_HSCROLL:  // slider control messages...

         pos = SendDlgItemMessage(hWnd, IDC_SLIDER1, TBM_GETPOS, 0, (LPARAM)0);
         wsprintf(szPos, "%01d.%01d", pos/10, pos % 10);
         SendDlgItemMessage(hWnd, IDC_EDIT2, WM_SETTEXT, 0, (LPARAM)((LPCSTR)szPos));

         break;

      case WM_CLOSE:

         SendMessage(hWnd, WM_COMMAND, IDCLOSE, 0L);

         break;
   }

   return FALSE;
}
